# Copyright (C) 2026 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import datetime as dt

from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.protocol import Node
from nbxmpp.structs import AccountInviteResult


def parse_account_invite(form: Node) -> AccountInviteResult:
    form = SimpleDataForm(extend=form)
    assert isinstance(form, SimpleDataForm)
    # TODO: Check form type once its standardized
    # field = form.vars.get("FORM_TYPE")
    # if field is None:
    #     return None

    try:
        uri = form["uri"].value
    except Exception:
        raise ValueError("uri field is missing")

    assert isinstance(uri, str)
    if not uri:
        raise ValueError("empty uri value")

    try:
        landing_url = form["landing-url"].value or None
    except Exception:
        landing_url = None

    assert isinstance(landing_url, str | None)

    try:
        expire = parse_datetime(form["expire"].value)
    except Exception:
        expire = None

    assert isinstance(expire, dt.datetime | None)

    return AccountInviteResult(
        uri=uri,
        landing_url=landing_url,
        expire=expire,
    )



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import isMucPM
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import ReceiptData
from nbxmpp.structs import StanzaHandler
from nbxmpp.util import generate_id

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Receipts(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_receipt,
                ns=Namespace.RECEIPTS,
                priority=15,
            ),
        ]

    def _process_message_receipt(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        request = stanza.getTag("request", namespace=Namespace.RECEIPTS)
        if request is not None:
            properties.receipt = ReceiptData(request.getName())
            return

        received = stanza.getTag("received", namespace=Namespace.RECEIPTS)
        if received is not None:
            id_ = received.getAttr("id")
            if id_ is None:
                self._log.warning("Receipt without id attr")
                self._log.warning(stanza)
                return

            properties.receipt = ReceiptData(received.getName(), id_)


def build_receipt(stanza: Message | Any) -> Message:
    if not isinstance(stanza, Message):
        raise ValueError("Stanza type must be protocol.Message")

    if stanza.getType() == "error":
        raise ValueError("Receipt can not be generated for type error messages")

    if stanza.getID() is None:
        raise ValueError("Receipt can not be generated for messages without id")

    if stanza.getTag("received", namespace=Namespace.RECEIPTS) is not None:
        raise ValueError("Receipt can not be generated for receipts")

    is_muc_pm = isMucPM(stanza)

    jid = stanza.getFrom()
    typ = stanza.getType()
    if typ == "groupchat" or not is_muc_pm:
        jid = jid.new_as_bare()

    message = Message(to=jid, typ=typ)
    if is_muc_pm:
        message.setTag("x", namespace=Namespace.MUC_USER)
    message_id = generate_id()
    message.setID(message_id)
    message.setReceiptReceived(stanza.getID())
    message.setHint("store")
    message.setOriginID(message_id)
    return message



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import MessageType
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.fallback import parse_fallback_indication
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import BodyData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import StanzaIDData
from nbxmpp.structs import XHTMLData
from nbxmpp.util import error_factory

if TYPE_CHECKING:
    from nbxmpp.client import Client


class BaseMessage(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message", callback=self._process_message_base, priority=5
            ),
            StanzaHandler(
                name="message", callback=self._process_message_after_base, priority=10
            ),
        ]

    def _process_message_base(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        properties.type = self._parse_type(stanza)

        if properties.is_carbon_message and properties.carbon.is_sent:
            properties.jid = stanza.getTo()

        elif properties.is_mam_message and not properties.type.is_groupchat:
            own_jid = self._client.get_bound_jid()
            if own_jid.bare_match(stanza.getFrom()):
                properties.jid = stanza.getTo()
            else:
                properties.jid = stanza.getFrom()

        else:
            properties.jid = stanza.getFrom()

        self._parse_if_private_message(stanza, properties)

        properties.remote_jid = self._determine_remote_jid(properties)
        properties.from_ = stanza.getFrom()
        properties.to = stanza.getTo()
        properties.id = stanza.getID()
        properties.self_message = self._parse_self_message(stanza, properties)

        properties.origin_id = stanza.getOriginID()
        properties.stanza_ids = self._parse_stanza_ids(stanza)

        if properties.type.is_error:
            properties.error = error_factory(stanza)

    def _determine_remote_jid(self, properties: MessageProperties) -> JID | None:
        if properties.is_muc_pm:
            return properties.jid
        return properties.jid.new_as_bare()

    def _parse_if_private_message(
        self, stanza: Message, properties: MessageProperties
    ) -> None:
        muc_user = stanza.getTag("x", namespace=Namespace.MUC_USER)
        if muc_user is None:
            return

        if not properties.jid.is_full:
            return

        if properties.type.is_chat or (
            properties.type.is_error and not muc_user.getChildren()
        ):
            properties.muc_private_message = True

    def _process_message_after_base(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        # This handler runs after decryption, when handlers had the chance
        # to decrypt the body

        fallbacks_for = parse_fallback_indication(self._log, stanza)

        properties.body = stanza.getBody()
        properties.bodies = BodyData(
            stanza, fallbacks_for, self._client.get_supported_fallback_ns()
        )
        properties.thread = stanza.getThread()
        properties.subject = stanza.getSubject()
        forms = stanza.getTags("x", namespace=Namespace.DATA)
        if forms:
            properties.forms = forms

        xhtml = stanza.getXHTML()
        if xhtml is None:
            return

        if xhtml.getTag("body", namespace=Namespace.XHTML) is None:
            self._log.warning("xhtml without body found")
            self._log.warning(stanza)
            return

        properties.xhtml = XHTMLData(xhtml)

    def _parse_type(self, stanza: Message) -> MessageType:
        type_ = stanza.getType()
        if type_ is None:
            return MessageType.NORMAL

        try:
            return MessageType(type_)
        except ValueError:
            self._log.warning("Message with invalid type: %s", type_)
            self._log.warning(stanza)
            raise NodeProcessed

    @staticmethod
    def _parse_self_message(stanza: Message, properties: MessageProperties) -> bool:
        if properties.type.is_groupchat:
            return False
        return stanza.getFrom().bare_match(stanza.getTo())

    def _parse_stanza_ids(self, stanza: Message) -> list[StanzaIDData]:
        stanza_ids: list[StanzaIDData] = []
        for stanza_id in stanza.getTags("stanza-id", namespace=Namespace.SID):
            id_ = stanza_id.getAttr("id")
            by = stanza_id.getAttr("by")
            if not id_ or not by:
                self._log.warning("Missing attributes on stanza-id")
                self._log.warning(stanza)
                continue

            stanza_ids.append(StanzaIDData(id=id_, by=by))

        return stanza_ids



# Copyright (C) 2021 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import datetime as dt
from collections.abc import Callable

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import create_tzinfo
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_FORBIDDEN
from nbxmpp.protocol import ERR_SERVICE_UNAVAILABLE
from nbxmpp.protocol import Error
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class EntityTime(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._answer_request,
                priority=60,
                typ="get",
                ns=Namespace.TIME,
            ),
        ]

        self._enabled = False
        self._allow_reply_func: Callable[..., bool] | None = None

    def disable(self) -> None:
        self._enabled = False

    def enable(self) -> None:
        self._enabled = True

    def set_allow_reply_func(self, func: Callable[..., bool]) -> None:
        self._allow_reply_func = func

    @iq_request_task
    def request_entity_time(self, jid: JID):
        _task = yield

        response = yield _make_request(jid)
        if response.isError():
            raise StanzaError(response)

        yield _parse_response(response)

    def _answer_request(
        self, _client: Client, stanza: Iq, _properties: IqProperties
    ) -> None:
        self._log.info("Request received from %s", stanza.getFrom())
        if not self._enabled:
            self._client.send_stanza(Error(stanza, ERR_SERVICE_UNAVAILABLE))
            raise NodeProcessed

        if self._allow_reply_func is not None:
            if not self._allow_reply_func(stanza.getFrom()):
                self._client.send_stanza(Error(stanza, ERR_FORBIDDEN))
                raise NodeProcessed

        time, tzo = get_local_time()
        iq = stanza.buildSimpleReply("result")
        time_node = iq.addChild("time", namespace=Namespace.TIME)
        time_node.setTagData("utc", time)
        time_node.setTagData("tzo", tzo)
        self._log.info("Send time: %s %s", time, tzo)
        self._client.send_stanza(iq)
        raise NodeProcessed


def get_local_time() -> tuple[str, str]:
    utc = dt.datetime.now(tz=dt.timezone.utc)
    utc_formated = utc.strftime("%Y-%m-%dT%H:%M:%SZ")
    local = utc.astimezone()
    if local.utcoffset() == utc.utcoffset():
        offset = "Z"
    else:
        # %z has the format ±HHMM[SS[.ffffff]]
        offset = local.strftime("%z")[:5]
        offset = f"{offset[:-2]}:{offset[-2:]}"
    return utc_formated, offset


def _make_request(jid: JID) -> Iq:
    iq = Iq("get", to=jid)
    iq.addChild("time", namespace=Namespace.TIME)
    return iq


def _parse_response(response: Iq) -> str:
    time_ = response.getTag("time")
    if not time_:
        raise MalformedStanzaError("time node missing", response)

    tzo = time_.getTagData("tzo")
    if not tzo:
        raise MalformedStanzaError("tzo node or data missing", response)

    remote_tz = create_tzinfo(tz_string=tzo)
    if remote_tz is None:
        raise MalformedStanzaError("invalid tzo data", response)

    utc_time = time_.getTagData("utc")
    if not utc_time:
        raise MalformedStanzaError("utc node or data missing", response)

    date_time = parse_datetime(utc_time, check_utc=True)
    if date_time is None:
        raise MalformedStanzaError("invalid timezone definition", response)

    date_time = date_time.astimezone(remote_tz)
    return date_time.strftime("%c %Z")



# Copyright (C) 2022 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from dataclasses import fields

from nbxmpp import Node
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import OpenGraphData
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client
    from nbxmpp.protocol import Message


class OpenGraph(BaseModule):
    def __init__(self, client: Client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_opengraph,
                ns=Namespace.RDF,
                priority=15,
            ),
        ]

    def _process_message_opengraph(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        for desc in stanza.getTags("Description", namespace=Namespace.RDF):
            about = desc.getAttr("about") or desc.getNamespacedAttr(
                "about", Namespace.RDF
            )
            if about:
                properties.open_graph[about] = self._process_description(desc)

    @staticmethod
    def _process_description(description: Node) -> OpenGraphData:
        data = OpenGraphData()
        for field in fields(OpenGraphData):
            tag = field.name
            if tag == "image":
                for node in description.getTags(tag, namespace=Namespace.OPEN_GRAPH):
                    node_data = node.getData()
                    if node_data.startswith("data:image"):
                        setattr(data, tag, node_data)
                        break

            else:
                if node := description.getTag(tag, namespace=Namespace.OPEN_GRAPH):
                    setattr(data, tag, node.getData())

        return data



import unittest

from nbxmpp.modules.delay import parse_delay
from nbxmpp.protocol import Node


class TestHelpers(unittest.TestCase):

    def test_parse_delay(self):

        node = """
        <message>
            <delay xmlns='urn:xmpp:delay' from='capulet.com' stamp='2002-09-10T23:08:25Z' />
            <delay xmlns='urn:xmpp:delay' from='romeo.com' stamp='2010-09-10T23:08:25Z' />
            <delay xmlns='urn:xmpp:delay' stamp='2015-09-10T23:08:25Z' />
        </message>
        """
        message = Node(node=node)

        timestamp = parse_delay(message)
        self.assertEqual(timestamp, 1031699305.0)

        timestamp = parse_delay(message, from_=["capulet.com"])
        self.assertEqual(timestamp, 1031699305.0)

        timestamp = parse_delay(message, from_=["romeo.com"])
        self.assertEqual(timestamp, 1284160105.0)

        timestamp = parse_delay(message, not_from=["romeo.com"])
        self.assertEqual(timestamp, 1031699305.0)


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import ACTIVITIES
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import ActivityData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Activity(BaseModule):

    _depends = {"publish": "PubSub"}

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_activity,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
        ]

    def _process_pubsub_activity(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.ACTIVITY:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        activity_node = item.getTag("activity", namespace=Namespace.ACTIVITY)
        if not activity_node.getChildren():
            self._log.info("Received activity: %s - no activity set", properties.jid)
            return

        activity, subactivity, text = None, None, None
        for child in activity_node.getChildren():
            name = child.getName()
            if name == "text":
                text = child.getData()
            elif name in ACTIVITIES:
                activity = name
                subactivity = self._parse_sub_activity(child)

        if activity is None and activity_node.getChildren():
            self._log.warning("No valid activity value found")
            self._log.warning(stanza)
            raise NodeProcessed

        data = ActivityData(activity, subactivity, text)
        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info("Received activity: %s - %s", properties.jid, data)

        properties.pubsub_event = pubsub_event

    @staticmethod
    def _parse_sub_activity(activity: Node) -> str | None:
        sub_activities = ACTIVITIES[activity.getName()]
        for sub in activity.getChildren():
            if sub.getName() in sub_activities:
                return sub.getName()
        return None

    @iq_request_task
    def set_activity(self, data: ActivityData):
        task = yield

        item = Node("activity", {"xmlns": Namespace.ACTIVITY})
        if data is not None and data.activity:
            activity_node = item.addChild(data.activity)
            if data.subactivity:
                activity_node.addChild(data.subactivity)
            if data.text:
                item.addChild("text", payload=data.text)

        result = yield self.publish(Namespace.ACTIVITY, item, id_="current")

        yield finalize(task, result)



#!/usr/bin/env python3

import argparse
import re
import subprocess
import sys
from pathlib import Path

REPO_DIR = Path(__file__).resolve().parent.parent


INIT = REPO_DIR / "nbxmpp" / "__init__.py"
CHANGELOG = REPO_DIR / "ChangeLog"

VERSION_RX = r"\d+\.\d+\.\d+"


def get_current_version() -> str:
    content = INIT.read_text(encoding="utf8")
    match = re.search(VERSION_RX, content)
    if match is None:
        sys.exit("Unable to find current version")
    return match[0]


def bump_version(current_version: str, new_version: str) -> None:
    content = INIT.read_text(encoding="utf8")
    content = content.replace(current_version, new_version, 1)
    INIT.write_text(content, encoding="utf8")


def make_changelog(new_version: str) -> None:

    cmd = ["git-chglog", "--next-tag", new_version]

    result = subprocess.run(  # noqa: S603
        cmd, cwd=REPO_DIR, text=True, check=True, capture_output=True
    )

    changes = result.stdout
    changes = changes.removeprefix("\n")

    current_changelog = CHANGELOG.read_text()

    with CHANGELOG.open("w") as f:
        f.write(changes + current_changelog)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Bump Version")
    parser.add_argument("version", help="The new version, e.g. 1.5.0")
    args = parser.parse_args()

    current_version = get_current_version()

    bump_version(current_version, args.version)
    make_changelog(args.version)



from test.lib.util import StanzaHandlerTest

from nbxmpp.modules.reactions import Reactions
from nbxmpp.protocol import Message
from nbxmpp.structs import Reactions as ReactionStruct


class MockLog:
    @staticmethod
    def warning(_):
        pass


class MockModule:
    _log = MockLog

    @staticmethod
    def is_emoji(s):
        return Reactions.is_emoji(s)


class ReactionsTest(StanzaHandlerTest):
    def test_reaction_parsing(self):
        class P:
            reactions: ReactionStruct

        xml = """
            <message to='romeo@capulet.net/orchard' id='96d73204-a57a-11e9-88b8-4889e7820c76' type='chat'>
              <reactions id='744f6e18-a57a-11e9-a656-4889e7820c76' xmlns='urn:xmpp:reactions:0'>
                <reaction>👋  </reaction>
                <reaction>🐢</reaction>
              </reactions>
              <store xmlns='urn:xmpp:hints'/>
            </message>
        """
        msg = Message(node=xml)
        Reactions._process_message_reaction(MockModule, self, msg, P)

        self.assertEqual(P.reactions.id, "744f6e18-a57a-11e9-a656-4889e7820c76")
        self.assertEqual(P.reactions.emojis, {"👋", "🐢"})

    def test_no_reactions(self):
        class P:
            reactions: ReactionStruct = None

        xml = """
            <message to='romeo@capulet.net/orchard' id='96d73204-a57a-11e9-88b8-4889e7820c76' type='chat'>
              <store xmlns='urn:xmpp:hints'/>
            </message>
        """
        msg = Message(node=xml)
        Reactions._process_message_reaction(MockModule, self, msg, P)

        self.assertIsNone(P.reactions)

    def test_empty_reactions(self):
        class P:
            reactions: ReactionStruct

        xml = """
            <message to='romeo@capulet.net/orchard' id='96d73204-a57a-11e9-88b8-4889e7820c76' type='chat'>
              <reactions id='744f6e18-a57a-11e9-a656-4889e7820c76' xmlns='urn:xmpp:reactions:0' />
              <store xmlns='urn:xmpp:hints'/>
            </message>
        """
        msg = Message(node=xml)
        Reactions._process_message_reaction(MockModule, self, msg, P)

        self.assertEqual(len(P.reactions.emojis), 0)

    def test_invalid_reactions_no_id(self):
        class P:
            reactions: ReactionStruct

        xml = """
            <message to='romeo@capulet.net/orchard' id='96d73204-a57a-11e9-88b8-4889e7820c76' type='chat'>
              <reactions xmlns='urn:xmpp:reactions:0'>
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
              <store xmlns='urn:xmpp:hints'/>
            </message>
        """
        msg = Message(node=xml)
        Reactions._process_message_reaction(MockModule, self, msg, P)
        self.assertFalse(hasattr(P, "reactions"))

    def test_invalid_reactions_empty_id(self):
        class P:
            reactions: ReactionStruct

        xml = """
            <message to='romeo@capulet.net/orchard' id='96d73204-a57a-11e9-88b8-4889e7820c76' type='chat'>
              <reactions id='' xmlns='urn:xmpp:reactions:0'>
                <reaction>👋</reaction>
                <reaction>🐢</reaction>
              </reactions>
              <store xmlns='urn:xmpp:hints'/>
            </message>
        """
        msg = Message(node=xml)
        Reactions._process_message_reaction(MockModule, self, msg, P)
        self.assertFalse(hasattr(P, "reactions"))

    def test_invalid_reactions_empty_emoji(self):
        class P:
            reactions: ReactionStruct

        xml = """
            <message to='romeo@capulet.net/orchard' id='96d73204-a57a-11e9-88b8-4889e7820c76' type='chat'>
              <reactions id='sadfsadf' xmlns='urn:xmpp:reactions:0'>
                <reaction></reaction>
                <reaction>🐢</reaction>
              </reactions>
              <store xmlns='urn:xmpp:hints'/>
            </message>
        """
        msg = Message(node=xml)
        Reactions._process_message_reaction(MockModule, self, msg, P)
        self.assertEqual(P.reactions.emojis, {"🐢"})

    def test_set_reactions(self):
        x = Message()
        x.setReactions("id", "🐢")
        self.assertEqual(x.getReactions(), ("id", {"🐢"}))

        x = Message()
        x.setReactions("id", "🐢👋")
        self.assertEqual(x.getReactions(), ("id", {"🐢", "👋"}))

        x = Message()
        x.setReactions("id", "")
        self.assertEqual(x.getReactions(), ("id", set()))

        x = Message()
        self.assertIsNone(x.getReactions())



import unittest

from nbxmpp.modules.discovery import parse_disco_info
from nbxmpp.protocol import Iq


class RestrictedReactions(unittest.TestCase):
    def test_form_present(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
              <field var='os'>
                <value>Mac</value>
              </field>
              <field var='os_version'>
                <value>10.5.1</value>
              </field>
              <field var='software'>
                <value>Psi</value>
              </field>
              <field var='software_version'>
                <value>0.11</value>
              </field>
            </x>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:reactions:0:restrictions</value>
              </field>
              <field var='max_reactions_per_user'>
                <value>1</value>
              </field>
              <field var='allowlist' type='list-multi'>
                <value>💘</value>
                <value>❤️</value>
                <value>💜</value>
              </field>
            </x>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        self.assertEqual(info.reactions_per_user, 1)
        self.assertEqual(info.reactions_allowlist, ["💘", "❤️", "💜"])

    def test_form_absent(self):
        node = """
        <iq from='benvolio@capulet.lit/230193' id='disco1' to='juliet@capulet.lit/chamber' type='result'>
          <query xmlns='http://jabber.org/protocol/disco#info' node='http://psi-im.org#q07IKJEyjvHSyhy//CH0CxmKi8w='>
            <identity xml:lang='en' category='client' name='Psi 0.11' type='pc'/>
            <identity xml:lang='el' category='client' name='Ψ 0.11' type='pc'/>
            <feature var='http://jabber.org/protocol/caps'/>
            <feature var='http://jabber.org/protocol/disco#info'/>
            <feature var='http://jabber.org/protocol/disco#items'/>
            <feature var='http://jabber.org/protocol/muc'/>
            <x xmlns='jabber:x:data' type='result'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:dataforms:softwareinfo</value>
              </field>
              <field var='ip_version'>
                <value>ipv4</value>
                <value>ipv6</value>
              </field>
              <field var='os'>
                <value>Mac</value>
              </field>
              <field var='os_version'>
                <value>10.5.1</value>
              </field>
              <field var='software'>
                <value>Psi</value>
              </field>
              <field var='software_version'>
                <value>0.11</value>
              </field>
            </x>
          </query>
        </iq>"""

        info = parse_disco_info(Iq(node=node))
        self.assertIsNone(info.reactions_per_user)
        self.assertIsNone(info.reactions_allowlist)



import unittest

from nbxmpp.simplexml import Node


class TestNode(unittest.TestCase):

    def test_topretty(self):
        string = """<presence xmlns="jabber:client" xml:lang="en" to="somejid@jid.com">
  <c xmlns="http://jabber.org/protocol/caps" ver="iVeWK58IHqW8e1wc9u4OGClblVo=" />
  <x xmlns="vcard-temp:x:update">
    <photo>bb055be7076edf87c9f89e4e0b829f0624aa1cef</photo>
  </x>
  <occupant-id xmlns="urn:xmpp:occupant-id:0" id="yAkJlge8v5CxmIAXT1m3jwuXLcRWidERkCGco9XN5z0=" />
  <x xmlns="http://jabber.org/protocol/muc#user">
    <item affiliation="none" role="participant" />
  </x>
  <priority>127</priority>
</presence>"""

        node = Node(node=string)
        self.assertEqual(node.topretty(), string)

        string = """<a xmlns="http://www.gajim.org/xmlns/undeclared">
  abc
  <b />
  <c />
  fgh
</a>"""

        node = Node(node=string)
        self.assertEqual(node.topretty(), string)

        string = """<a xmlns="http://www.gajim.org/xmlns/undeclared">
  <b />
  abc
  <c />
  fgh
</a>"""

        node = Node(node=string)
        self.assertEqual(node.topretty(), string)

        string = """<a xmlns="http://www.gajim.org/xmlns/undeclared">
  <b />
  abc
  <c>123</c>
  fgh
</a>"""

        node = Node(node=string)
        self.assertEqual(node.topretty(), string)

        string = """<a xmlns="http://www.gajim.org/xmlns/undeclared">
  <b>
    <c>
      <d />
      <e />
      <f />
    </c>
  </b>
</a>"""

        node = Node(node=string)
        self.assertEqual(node.topretty(), string)


if __name__ == "__main__":
    unittest.main()



import unittest

from nbxmpp.modules.vcard4 import LanguageParameter
from nbxmpp.modules.vcard4 import VCard
from nbxmpp.simplexml import Node


class TestVCard4(unittest.TestCase):

    def test_vard4_parsing(self):
        vcard_node = Node(
            node="""
            <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                <fn><text>Peter Saint-Andre</text></fn>
                <n>
                  <parameters>
                    <language><language-tag>fr</language-tag></language>
                  </parameters>
                  <surname>Saint-Andre</surname>
                  <given>Peter</given>
                  <additional></additional>
                </n>
                <nickname><text>stpeter</text></nickname>
                <nickname><text>psa</text></nickname>
                <photo><uri>https://stpeter.im/images/stpeter_oscon.jpg</uri></photo>
                <bday><date>1966-08-06</date></bday>
                <adr>
                  <parameters>
                    <type><text>work</text><text>voice</text></type>
                    <pref><integer>1</integer></pref>
                  </parameters>
                  <ext>Suite 600</ext>
                  <street>1899 Wynkoop Street</street>
                  <locality>Denver</locality>
                  <region>CO</region>
                  <code>80202</code>
                  <country>USA</country>
                </adr>
                <adr>
                  <parameters><type><text>home</text></type></parameters>
                  <ext></ext>
                  <street></street>
                  <locality>Parker</locality>
                  <region>CO</region>
                  <code>80138</code>
                  <country>USA</country>
                </adr>
                <tel>
                  <parameters>
                    <type><text>work</text><text>voice</text></type>
                    <pref><integer>1</integer></pref>
                  </parameters>
                  <uri>tel:+1-303-308-3282</uri>
                </tel>
                <tel>
                  <parameters><type><text>work</text><text>fax</text></type></parameters>
                  <uri>tel:+1-303-308-3219</uri>
                </tel>
                <tel>
                  <parameters>
                    <type><text>cell</text><text>voice</text><text>text</text></type>
                  </parameters>
                  <uri>tel:+1-720-256-6756</uri>
                </tel>
                <tel>
                  <parameters><type><text>home</text><text>voice</text></type></parameters>
                  <uri>tel:+1-303-555-1212</uri>
                </tel>
                <geo><uri>geo:39.59,-105.01</uri></geo>
                <title><text>Executive Director</text></title>
                <role><text>Patron Saint</text></role>
                <org>
                  <parameters><type><text>work</text></type></parameters>
                  <text>XMPP Standards Foundation</text>
                </org>
                <url><uri>https://stpeter.im/</uri></url>
                <note>
                  <text>
                  More information about me is located on my
                  personal website: https://stpeter.im/
                  </text>
                </note>
                <gender><sex><text>M</text></sex></gender>
                <lang>
                  <parameters><pref>1</pref></parameters>
                  <language-tag>en</language-tag>
                </lang>
                <email>
                  <parameters><type><text>work</text></type></parameters>
                  <text>psaintan@cisco.com</text>
                </email>
                <email>
                  <parameters><type><text>home</text></type></parameters>
                  <text>stpeter@jabber.org</text>
                </email>
                <impp>
                  <parameters><type><text>work</text></type></parameters>
                  <uri>xmpp:psaintan@cisco.com</uri>
                </impp>
                <impp>
                  <parameters><type><text>home</text></type></parameters>
                  <uri>xmpp:stpeter@jabber.org</uri>
                </impp>
                <key>
                  <uri>https://stpeter.im/stpeter.asc</uri>
                </key>
                <pronouns>
                  <text>he/him</text>
                </pronouns>
                <unsupported-element>unsupported</unsupported-element>
            </vcard>
        """
        )

        vcard = VCard.from_node(vcard_node)
        props = vcard.get_properties()

        email_props = list(filter(lambda p: p.name == "email", props))
        self.assertEqual(len(email_props), 2)

        nickname_props = list(filter(lambda p: p.name == "nickname", props))
        self.assertEqual(len(nickname_props), 2)

        n_props = list(filter(lambda p: p.name == "n", props))
        self.assertEqual(len(n_props), 1)
        n_prop = n_props[0]
        lang_param = n_prop.parameters.get_parameter("language")
        assert isinstance(lang_param, LanguageParameter)
        self.assertEqual(lang_param.value, "fr")

        pronouns_props = list(filter(lambda p: p.name == "pronouns", props))
        self.assertEqual(len(pronouns_props), 1)
        self.assertEqual(pronouns_props[0].value, "he/him")

        # Preserve unsupported elements
        node = vcard.to_node()
        self.assertEqual(node.getTagData("unsupported-element"), "unsupported")

    def test_vcard4_pronouns(self):
        vcard_node = Node(
            node="""
            <vcard xmlns="urn:ietf:params:xml:ns:vcard-4.0">
                <fn><text>Cartman</text></fn>
                <pronouns>
                  <parameters>
                    <language><language-tag>fr</language-tag></language>
                    <pref>
                      <integer>1</integer>
                    </pref>
                  </parameters>
                  <text>il/lui</text>
                </pronouns>
                <pronouns>
                  <text>he/him</text>
                </pronouns>
            </vcard>
        """
        )

        vcard = VCard.from_node(vcard_node)
        props = vcard.get_properties()
        pronouns_props = list(filter(lambda p: p.name == "pronouns", props))
        self.assertEqual(len(pronouns_props), 2)
        self.assertEqual(
            pronouns_props[0].parameters.get_parameter("language").value, "fr"
        )



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.bookmarks.util import build_storage_node
from nbxmpp.modules.bookmarks.util import get_private_request
from nbxmpp.modules.bookmarks.util import parse_private_bookmarks
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.structs import BookmarkData
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class PrivateBookmarks(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_bookmarks(self):
        _task = yield

        response = yield get_private_request()
        if response.isError():
            raise StanzaError(response)

        bookmarks = parse_private_bookmarks(response, self._log)
        for bookmark in bookmarks:
            self._log.info(bookmark)

        yield bookmarks

    @iq_request_task
    def store_bookmarks(self, bookmarks: list[BookmarkData]):
        _task = yield

        self._log.info("Store Bookmarks")

        storage_node = build_storage_node(bookmarks)
        response = yield Iq("set", Namespace.PRIVATE, payload=storage_node)
        yield process_response(response)



import unittest
from unittest.mock import Mock

from nbxmpp.sasl import SCRAM_SHA_1
from nbxmpp.util import b64encode

# Test vector from https://wiki.xmpp.org/web/SASL_and_SCRAM-SHA-1


class SCRAM(unittest.TestCase):
    def setUp(self):
        self.con = Mock()
        self.maxDiff = None
        self._username = "user"
        self._password = "pencil"  # noqa: S105
        self._mechanism = SCRAM_SHA_1(self._username, self._password, None)
        self._mechanism._client_nonce = "fyko+d2lbbFgONRv9qkxdawL"

    def test_auth(self):
        initial = b64encode("n,,n=user,r=fyko+d2lbbFgONRv9qkxdawL")
        data = self._mechanism.get_initiate_data()
        self.assertEqual(data, initial)

        challenge = b64encode(
            "r=fyko+d2lbbFgONRv9qkxdawL3rfcNHYJY1ZVvWVs7j,s=QSXCR+Q6sek8bf92,i=4096"
        )
        data = self._mechanism.get_response_data(challenge)

        response = b64encode(
            "c=biws,r=fyko+d2lbbFgONRv9qkxdawL3rfcNHYJY1ZVvWVs7j,p=v0X8v3Bz2T0CJGbJQyF0X+HI4Ts="
        )
        self.assertEqual(data, response)

        success = b64encode("v=rmF9pqV8S7suAoZWja4dJRkFsKQ=")
        self._mechanism.validate_success_data(success)


if __name__ == "__main__":
    unittest.main()



from test.lib.util import StanzaHandlerTest

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import TuneData


class TuneTest(StanzaHandlerTest):

    def test_tune_parsing(self):
        def _on_message(_con, _stanza, properties):

            data = TuneData(
                artist="Yes",
                length="686",
                rating="8",
                source="Yessongs",
                title="Heart of the Sunrise",
                track="3",
                uri="https://www.artist.com",
            )

            pubsub_event = PubSubEventData(
                node="http://jabber.org/protocol/tune",
                id="bffe6584-0f9c-11dc-84ba-001143d5d5db",
                item=None,
                data=data,
                deleted=False,
                retracted=False,
                purged=False,
            )

            # We cant compare Node objects
            pubsub_event_ = properties.pubsub_event._replace(item=None)
            self.assertEqual(pubsub_event, pubsub_event_)

        event = """
            <message from='test@test.test'>
                <event xmlns='http://jabber.org/protocol/pubsub#event'>
                    <items node='http://jabber.org/protocol/tune'>
                        <item id='bffe6584-0f9c-11dc-84ba-001143d5d5db'>
                            <tune xmlns='http://jabber.org/protocol/tune'>
                                <artist>Yes</artist>
                                <length>686</length>
                                <rating>8</rating>
                                <source>Yessongs</source>
                                <title>Heart of the Sunrise</title>
                                <track>3</track>
                                <uri>https://www.artist.com</uri>
                            </tune>
                        </item>
                    </items>
                </event>
            </message>
        """

        self.dispatcher.register_handler(
            StanzaHandler(
                name="message", callback=_on_message, ns=Namespace.PUBSUB_EVENT
            )
        )

        self.dispatcher.process_data(event)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_BAD_REQUEST
from nbxmpp.protocol import ERR_FEATURE_NOT_IMPLEMENTED
from nbxmpp.protocol import Error as ErrorStanza
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IBBData
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode

if TYPE_CHECKING:
    from nbxmpp.client import Client


class IBB(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq", callback=self._process_ibb, ns=Namespace.IBB, priority=20
            ),
        ]

    def _process_ibb(
        self, _client: Client, stanza: Iq, properties: IqProperties
    ) -> None:
        if properties.type.is_set:
            open_ = stanza.getTag("open", namespace=Namespace.IBB)
            if open_ is not None:
                properties.ibb = self._parse_open(stanza, open_)
                return

            close = stanza.getTag("close", namespace=Namespace.IBB)
            if close is not None:
                properties.ibb = self._parse_close(stanza, close)
                return

            data = stanza.getTag("data", namespace=Namespace.IBB)
            if data is not None:
                properties.ibb = self._parse_data(stanza, data)
                return

    def _parse_open(self, stanza: Iq, open_: Node) -> IBBData:
        attrs = open_.getAttrs()
        try:
            block_size = int(attrs.get("block-size"))
        except Exception as error:
            self._log.warning(error)
            self._log.warning(stanza)
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        if block_size > 65535:
            self._log.warning("Invalid block-size")
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        sid = attrs.get("sid")
        if not sid:
            self._log.warning("Invalid sid")
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        type_ = attrs.get("stanza")
        if type_ == "message":
            self._client.send_stanza(ErrorStanza(stanza, ERR_FEATURE_NOT_IMPLEMENTED))
            raise NodeProcessed

        return IBBData(type="open", block_size=block_size, sid=sid)

    def _parse_close(self, stanza: Iq, close: Node) -> IBBData:
        sid = close.getAttrs().get("sid")
        if sid is None:
            self._log.warning("Invalid sid")
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed
        return IBBData(type="close", sid=sid)

    def _parse_data(self, stanza: Iq, data: Node) -> IBBData:
        attrs = data.getAttrs()

        sid = attrs.get("sid")
        if sid is None:
            self._log.warning("Invalid sid")
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        try:
            seq = int(attrs.get("seq"))
        except Exception:
            self._log.exception("Invalid seq")
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        try:
            decoded_data = b64decode(data.getData())
        except Exception:
            self._log.exception("Failed to decode IBB data")
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        return IBBData(type="data", sid=sid, seq=seq, data=decoded_data)

    def send_reply(self, stanza: Iq, error: str | None = None) -> None:
        if error is None:
            reply = stanza.buildReply("result")
            reply.getChildren().clear()
        else:
            reply = ErrorStanza(stanza, error)
        self._client.send_stanza(reply)

    @iq_request_task
    def send_open(self, jid: str, sid: str, block_size: int):
        _task = yield

        response = yield _make_ibb_open(jid, sid, block_size)
        yield process_response(response)

    @iq_request_task
    def send_close(self, jid: str, sid: str):
        _task = yield

        response = yield _make_ibb_close(jid, sid)
        yield process_response(response)

    @iq_request_task
    def send_data(self, jid: str, sid: str, seq: int, data: bytes):
        _task = yield

        response = yield _make_ibb_data(jid, sid, seq, data)
        yield process_response(response)


def _make_ibb_open(jid: str, sid: str, block_size: int) -> Iq:
    iq = Iq("set", to=jid)
    iq.addChild(
        "open",
        {"block-size": block_size, "sid": sid, "stanza": "iq"},
        namespace=Namespace.IBB,
    )
    return iq


def _make_ibb_close(jid: str, sid: str) -> Iq:
    iq = Iq("set", to=jid)
    iq.addChild("close", {"sid": sid}, namespace=Namespace.IBB)
    return iq


def _make_ibb_data(jid: str, sid: str, seq: int, data: bytes) -> Iq:
    iq = Iq("set", to=jid)
    ibb_data = iq.addChild("data", {"sid": sid, "seq": seq}, namespace=Namespace.IBB)
    ibb_data.setData(b64encode(data))
    return iq



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import logging
import time

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.dataforms import MultipleDataForm
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_ITEM_NOT_FOUND
from nbxmpp.protocol import ErrorNode
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import DiscoIdentity
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import DiscoItem
from nbxmpp.structs import DiscoItems
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

log = logging.getLogger("nbxmpp.m.discovery")


class Discovery(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="iq",
                callback=self._process_disco_info,
                typ="get",
                ns=Namespace.DISCO_INFO,
                priority=90,
            ),
        ]

    @staticmethod
    def _process_disco_info(
        client: Client, stanza: Iq, _properties: IqProperties
    ) -> None:
        iq = stanza.buildReply("error")
        iq.addChild(node=ErrorNode(ERR_ITEM_NOT_FOUND))
        client.send_stanza(iq)
        raise NodeProcessed

    @iq_request_task
    def disco_info(self, jid: JID | str, node: str | None = None):
        _task = yield

        response = yield get_disco_request(Namespace.DISCO_INFO, jid, node)
        if response.isError():
            raise StanzaError(response)
        yield parse_disco_info(response)

    @iq_request_task
    def disco_items(self, jid: JID | str, node: str | None = None):
        _task = yield

        response = yield get_disco_request(Namespace.DISCO_ITEMS, jid, node)
        if response.isError():
            raise StanzaError(response)
        yield parse_disco_items(response)


def parse_disco_info(stanza: Iq, timestamp: float | None = None) -> DiscoInfo:
    identities: list[DiscoIdentity] = []
    features: list[str] = []
    dataforms: list[SimpleDataForm | MultipleDataForm] = []

    if timestamp is None:
        timestamp = time.time()

    query = stanza.getQuery()
    for node in query.getTags("identity"):
        attrs = node.getAttrs()
        try:
            identities.append(
                DiscoIdentity(
                    category=attrs["category"],
                    type=attrs["type"],
                    name=attrs.get("name"),
                    lang=attrs.get("xml:lang"),
                )
            )
        except Exception:
            raise MalformedStanzaError("invalid attributes", stanza)

    for node in query.getTags("feature"):
        try:
            features.append(node.getAttr("var"))
        except Exception:
            raise MalformedStanzaError("invalid attributes", stanza)

    for node in query.getTags("x", namespace=Namespace.DATA):
        dataforms.append(extend_form(node))

    return DiscoInfo(
        stanza=stanza,
        identities=identities,
        features=features,
        dataforms=dataforms,
        timestamp=timestamp,
    )


def parse_disco_items(stanza: Iq) -> DiscoItems:
    items: list[DiscoItem] = []

    query = stanza.getQuery()
    for node in query.getTags("item"):
        attrs = node.getAttrs()
        try:
            items.append(
                DiscoItem(
                    jid=attrs["jid"], name=attrs.get("name"), node=attrs.get("node")
                )
            )
        except Exception:
            raise MalformedStanzaError("invalid attributes", stanza)

    return DiscoItems(jid=stanza.getFrom(), node=query.getAttr("node"), items=items)


def get_disco_request(namespace: str, jid: str, node: str | None = None) -> Iq:
    iq = Iq("get", to=jid, queryNS=namespace)
    if node:
        iq.getQuery().setAttr("node", node)
    return iq



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Literal
from typing import TYPE_CHECKING

from nbxmpp.const import InviteType
from nbxmpp.const import MessageType
from nbxmpp.const import StatusCode
from nbxmpp.errors import is_error
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.muc.util import build_direct_invite
from nbxmpp.modules.muc.util import build_mediated_invite
from nbxmpp.modules.muc.util import make_affiliation_request
from nbxmpp.modules.muc.util import make_cancel_config_request
from nbxmpp.modules.muc.util import make_captcha_request
from nbxmpp.modules.muc.util import make_config_request
from nbxmpp.modules.muc.util import make_destroy_request
from nbxmpp.modules.muc.util import make_set_affiliation_request
from nbxmpp.modules.muc.util import make_set_config_request
from nbxmpp.modules.muc.util import make_set_role_request
from nbxmpp.modules.muc.util import MucInfoResult
from nbxmpp.modules.muc.util import parse_muc_user
from nbxmpp.modules.util import parse_xmpp_uri
from nbxmpp.modules.util import process_response
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import DataField
from nbxmpp.protocol import DataForm
from nbxmpp.protocol import ERR_NOT_ACCEPTABLE
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.protocol import StanzaMalformed
from nbxmpp.structs import AffiliationResult
from nbxmpp.structs import DeclineData
from nbxmpp.structs import InviteData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import MucConfigResult
from nbxmpp.structs import MucDestroyed
from nbxmpp.structs import MucSubject
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import VoiceRequest
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

AffiliationT = Literal["owner", "admin", "member", "outcast", "none"]
RoleT = Literal["moderator", "participant", "visitor", "none"]


class MUC(BaseModule):

    _depends = {
        "disco_info": "Discovery",
        "request_vcard": "VCardTemp",
        "send_moderation_request": "Moderation",
    }

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="presence",
                callback=self._process_muc_presence,
                ns=Namespace.MUC,
                priority=11,
            ),
            StanzaHandler(
                name="presence",
                callback=self._process_muc_user_presence,
                ns=Namespace.MUC_USER,
                priority=11,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message_base,
                ns=Namespace.MUC_USER,
                priority=5,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message_before_decryption,
                typ="groupchat",
                priority=6,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message_after_decryption,
                typ="groupchat",
                priority=16,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_mediated_invite,
                typ="normal",
                ns=Namespace.MUC_USER,
                priority=11,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_direct_invite,
                typ="normal",
                ns=Namespace.CONFERENCE,
                priority=12,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_voice_request,
                typ="normal",
                ns=Namespace.DATA,
                priority=11,
            ),
            StanzaHandler(
                name="message",
                callback=self._process_message,
                ns=Namespace.MUC_USER,
                priority=13,
            ),
        ]

    @staticmethod
    def _process_muc_presence(
        _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        muc = stanza.getTag("x", namespace=Namespace.MUC)
        if muc is None:
            return
        properties.from_muc = True
        properties.muc_jid = properties.jid.new_as_bare()
        properties.muc_nickname = properties.jid.resource

        occupant_id = stanza.getTagAttr(
            "occupant-id", "id", namespace=Namespace.OCCUPANT_ID
        )

        properties.occupant_id = occupant_id

    def _process_muc_user_presence(
        self, _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        muc_user = stanza.getTag("x", namespace=Namespace.MUC_USER)
        if muc_user is None:
            return
        properties.from_muc = True
        properties.muc_jid = properties.jid.new_as_bare()

        occupant_id = stanza.getTagAttr(
            "occupant-id", "id", namespace=Namespace.OCCUPANT_ID
        )

        properties.occupant_id = occupant_id

        destroy = muc_user.getTag("destroy")
        if destroy is not None:
            alternate = destroy.getAttr("jid")
            if alternate is not None:
                try:
                    alternate = JID.from_string(alternate)
                except Exception as error:
                    self._log.warning("Invalid alternate JID provided: %s", error)
                    self._log.warning(stanza)
                    alternate = None
            properties.muc_destroyed = MucDestroyed(
                alternate=alternate,
                reason=muc_user.getTagData("reason"),
                password=muc_user.getTagData("password"),
            )
            return

        properties.muc_nickname = properties.jid.resource

        # https://xmpp.org/extensions/xep-0045.html#registrar-statuscodes
        message_status_codes = [
            StatusCode.NON_ANONYMOUS,
            StatusCode.SELF,
            StatusCode.CONFIG_ROOM_LOGGING,
            StatusCode.CREATED,
            StatusCode.NICKNAME_MODIFIED,
            StatusCode.REMOVED_BANNED,
            StatusCode.NICKNAME_CHANGE,
            StatusCode.REMOVED_KICKED,
            StatusCode.REMOVED_AFFILIATION_CHANGE,
            StatusCode.REMOVED_NONMEMBER_IN_MEMBERS_ONLY,
            StatusCode.REMOVED_SERVICE_SHUTDOWN,
            StatusCode.REMOVED_ERROR,
        ]

        codes: set[StatusCode] = set()
        for status in muc_user.getTags("status"):
            try:
                code = StatusCode(status.getAttr("code"))
            except ValueError:
                self._log.warning(
                    "Received invalid status code: %s", status.getAttr("code")
                )
                self._log.warning(stanza)
                continue
            if code in message_status_codes:
                codes.add(code)

        if codes:
            properties.muc_status_codes = codes

        try:
            properties.muc_user = parse_muc_user(muc_user)
        except StanzaMalformed as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        if (
            properties.muc_user is not None
            and properties.muc_user.role.is_none
            and not properties.type.is_unavailable
        ):
            self._log.warning("Malformed Stanza")
            self._log.warning(stanza)
            raise NodeProcessed

    def _process_message_before_decryption(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        properties.from_muc = True
        properties.muc_jid = properties.jid.new_as_bare()
        properties.muc_nickname = properties.jid.resource

        occupant_id = stanza.getTagAttr(
            "occupant-id", "id", namespace=Namespace.OCCUPANT_ID
        )

        properties.occupant_id = occupant_id

        addresses = stanza.getTag("addresses", namespace=Namespace.ADDRESS)
        if addresses is not None:
            address = addresses.getTag("address", attrs={"type": "ofrom"})
            if address is not None:
                properties.muc_ofrom = JID.from_string(address.getAttr("jid"))

    def _process_message_after_decryption(
        self, _client: Client, _stanza: Message, properties: MessageProperties
    ) -> None:
        if properties.subject is None:
            return

        if properties.body is not None or properties.thread is not None:
            return

        properties.muc_subject = MucSubject(
            text=properties.subject,
            author=properties.muc_nickname,
            timestamp=properties.user_timestamp,
        )

    def _process_message_base(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        muc_user = stanza.getTag("x", namespace=Namespace.MUC_USER)
        if muc_user is None:
            return

        try:
            properties.muc_user = parse_muc_user(muc_user, is_presence=False)
        except StanzaMalformed as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

    def _process_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        muc_user = stanza.getTag("x", namespace=Namespace.MUC_USER)
        if muc_user is None:
            return

        occupant_id = stanza.getTagAttr(
            "occupant-id", "id", namespace=Namespace.OCCUPANT_ID
        )

        properties.occupant_id = occupant_id

        if properties.is_muc_invite_or_decline:
            return

        properties.from_muc = True
        properties.muc_jid = properties.jid.new_as_bare()

        if not properties.jid.is_bare:
            return

        # MUC Config change
        # https://xmpp.org/extensions/xep-0045.html#registrar-statuscodes
        message_status_codes = [
            StatusCode.SHOWING_UNAVAILABLE,
            StatusCode.NOT_SHOWING_UNAVAILABLE,
            StatusCode.CONFIG_NON_PRIVACY_RELATED,
            StatusCode.CONFIG_ROOM_LOGGING,
            StatusCode.CONFIG_NO_ROOM_LOGGING,
            StatusCode.CONFIG_NON_ANONYMOUS,
            StatusCode.CONFIG_SEMI_ANONYMOUS,
            StatusCode.CONFIG_FULL_ANONYMOUS,
        ]

        codes: set[StatusCode] = set()
        for status in muc_user.getTags("status"):
            try:
                code = StatusCode(status.getAttr("code"))
            except ValueError:
                self._log.warning(
                    "Received invalid status code: %s", status.getAttr("code")
                )
                self._log.warning(stanza)
                continue
            if code in message_status_codes:
                codes.add(code)

        if codes:
            properties.muc_status_codes = codes

    @staticmethod
    def _process_direct_invite(
        _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        direct = stanza.getTag("x", namespace=Namespace.CONFERENCE)
        if direct is None:
            return

        if stanza.getTag("x", namespace=Namespace.MUC_USER) is not None:
            # not a direct invite
            # See https://xmpp.org/extensions/xep-0045.html#example-57
            # read implementation notes
            return

        data = {}
        data["muc"] = JID.from_string(direct.getAttr("jid"))
        data["from_"] = properties.jid
        data["reason"] = direct.getAttr("reason")
        data["password"] = direct.getAttr("password")
        data["continued"] = direct.getAttr("continue") == "true"
        data["thread"] = direct.getAttr("thread")
        data["type"] = InviteType.DIRECT
        properties.muc_invite = InviteData(**data)

    @staticmethod
    def _process_mediated_invite(
        _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        muc_user = stanza.getTag("x", namespace=Namespace.MUC_USER)
        if muc_user is None:
            return

        if properties.type != MessageType.NORMAL:
            return

        properties.from_muc = True
        properties.muc_jid = properties.jid.new_as_bare()

        data = {}

        invite = muc_user.getTag("invite")
        if invite is not None:
            data["muc"] = properties.jid.new_as_bare()
            data["from_"] = JID.from_string(invite.getAttr("from"))
            data["reason"] = invite.getTagData("reason")
            data["password"] = muc_user.getTagData("password")
            data["type"] = InviteType.MEDIATED

            data["continued"] = False
            data["thread"] = None
            continue_ = invite.getTag("continue")
            if continue_ is not None:
                data["continued"] = True
                data["thread"] = continue_.getAttr("thread")
            properties.muc_invite = InviteData(**data)
            return

        decline = muc_user.getTag("decline")
        if decline is not None:
            data["muc"] = properties.jid.new_as_bare()
            data["from_"] = JID.from_string(decline.getAttr("from"))
            data["reason"] = decline.getTagData("reason")
            properties.muc_decline = DeclineData(**data)
            return

    def _process_voice_request(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        data_form = stanza.getTag("x", namespace=Namespace.DATA)
        if data_form is None:
            return

        data_form = extend_form(data_form)
        try:
            if data_form["FORM_TYPE"].value != Namespace.MUC_REQUEST:
                return
        except KeyError:
            return

        nick = data_form["muc#roomnick"].value

        try:
            jid = JID.from_string(data_form["muc#jid"].value)
        except Exception:
            self._log.warning("Invalid JID on voice request")
            self._log.warning(stanza)
            raise NodeProcessed

        properties.voice_request = VoiceRequest(jid=jid, nick=nick, form=data_form)
        properties.from_muc = True
        properties.muc_jid = properties.jid.new_as_bare()

    def approve_voice_request(self, muc_jid: JID, voice_request: VoiceRequest) -> str:
        form = voice_request.form
        form.type_ = "submit"
        form["muc#request_allow"].value = True
        return self._client.send_stanza(Message(to=muc_jid, payload=form))

    @iq_request_task
    def get_affiliation(self, jid: JID | str, affiliation: AffiliationT):
        _task = yield

        response = yield make_affiliation_request(jid, affiliation)
        if response.isError():
            raise StanzaError(response)

        room_jid = response.getFrom()
        query = response.getTag("query", namespace=Namespace.MUC_ADMIN)
        items = query.getTags("item")
        users_dict = {}
        for item in items:
            user_jid = item.getAttr("jid")
            if user_jid is None:
                self._log.warning("Received an affiliation item without JID: %s", item)
                self._log.warning(response)
                continue
            try:
                jid = JID.from_string(user_jid)
            except Exception as e:
                self._log.warning("'%s' is an invalid JID: %s", user_jid, e)
                continue

            users_dict[jid] = {}
            if item.has_attr("nick"):
                users_dict[jid]["nick"] = item.getAttr("nick")
            if item.has_attr("role"):
                users_dict[jid]["role"] = item.getAttr("role")
            reason = item.getTagData("reason")
            if reason:
                users_dict[jid]["reason"] = reason

        self._log.info("Affiliations received from %s: %s", room_jid, users_dict)

        yield AffiliationResult(jid=room_jid, users=users_dict)

    @iq_request_task
    def destroy(self, room_jid: JID, reason: str | None = None, jid: str | None = None):
        _task = yield

        response = yield make_destroy_request(room_jid, reason, jid)
        yield process_response(response)

    @iq_request_task
    def request_info(
        self, jid: JID | str, request_vcard: bool = True, allow_redirect: bool = False
    ):
        _task = yield

        redirected = False

        disco_info = yield self.disco_info(jid)
        if is_error(disco_info):
            error_response = disco_info

            if not allow_redirect:
                raise error_response

            if error_response.condition != "gone":
                raise error_response

            try:
                jid = parse_xmpp_uri(error_response.condition_data)[0]
            except Exception:
                raise error_response

            redirected = True
            disco_info = yield self.disco_info(jid)
            raise_if_error(disco_info)

        if not request_vcard or not disco_info.supports(Namespace.VCARD):
            yield MucInfoResult(info=disco_info, redirected=redirected)

        vcard = yield self.request_vcard(jid)
        if is_error(vcard):
            yield MucInfoResult(info=disco_info, redirected=redirected)

        yield MucInfoResult(info=disco_info, vcard=vcard, redirected=redirected)

    @iq_request_task
    def set_config(self, room_jid: JID, form):
        _task = yield

        response = yield make_set_config_request(room_jid, form)
        yield process_response(response)

    @iq_request_task
    def request_config(self, room_jid: JID):
        _task = yield

        response = yield make_config_request(room_jid)
        if response.isError():
            raise StanzaError(response)

        jid = response.getFrom()
        children = response.getQueryChildren()

        for form in children:
            if form.getNamespace() == Namespace.DATA:
                dataform = extend_form(node=form)
                self._log.info("Config form received for %s", jid)
                yield MucConfigResult(jid=jid, form=dataform)

        yield MucConfigResult(jid=jid)

    @iq_request_task
    def cancel_config(self, room_jid: JID):
        _task = yield

        response = yield make_cancel_config_request(room_jid)
        yield process_response(response)

    @iq_request_task
    def set_affiliation(self, room_jid: JID, users_dict: dict[JID, dict[str, str]]):
        _task = yield

        response = yield make_set_affiliation_request(room_jid, users_dict)
        yield process_response(response)

    @iq_request_task
    def set_role(
        self, room_jid: JID, nick: str, role: RoleT, reason: str | None = None
    ):
        _task = yield

        response = yield make_set_role_request(room_jid, nick, role, reason)
        yield process_response(response)

    def moderate_message(
        self, namespace: str, room_jid: JID, stanza_id: str, reason: str | None = None
    ) -> None:
        self.send_moderation_request(namespace, room_jid, stanza_id, reason)

    def set_subject(self, room_jid: JID, subject: MucSubject | None) -> str:
        message = Message(room_jid, typ="groupchat", subject=subject)
        self._log.info("Set subject for %s", room_jid)
        return self._client.send_stanza(message)

    def decline(self, room: JID | str, to: JID | str, reason: str | None = None) -> str:
        message = Message(to=room)
        muc_user = message.addChild("x", namespace=Namespace.MUC_USER)
        decline = muc_user.addChild("decline", attrs={"to": to})
        if reason:
            decline.setTagData("reason", reason)
        return self._client.send_stanza(message)

    def request_voice(self, room: JID) -> str:
        message = Message(to=room)
        xdata = DataForm(typ="submit")
        xdata.addChild(node=DataField(name="FORM_TYPE", value=Namespace.MUC_REQUEST))
        xdata.addChild(
            node=DataField(name="muc#role", value="participant", typ="text-single")
        )
        message.addChild(node=xdata)
        return self._client.send_stanza(message)

    def invite(
        self,
        room: JID,
        to: JID,
        password: str | None,
        reason: str | None = None,
        continue_: bool = False,
        type_: InviteType = InviteType.MEDIATED,
    ) -> str:
        if type_ == InviteType.DIRECT:
            invite = build_direct_invite(room, to, reason, password, continue_)
        else:
            invite = build_mediated_invite(room, to, reason, password, continue_)
        return self._client.send_stanza(invite)

    @iq_request_task
    def send_captcha(self, room_jid: JID, form_node):
        _task = yield

        response = yield make_captcha_request(room_jid, form_node)
        yield process_response(response)

    def cancel_captcha(self, room_jid: JID, message_id: str) -> None:
        message = Message(typ="error", to=room_jid)
        message.setID(message_id)
        message.setError(ERR_NOT_ACCEPTABLE)
        self._client.send_stanza(message)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from collections.abc import Iterator

from nbxmpp.const import ConnectionProtocol
from nbxmpp.const import ConnectionType
from nbxmpp.structs import ProxyData
from nbxmpp.structs import ServerAddress
from nbxmpp.types import CustomHostT
from nbxmpp.util import Observable
from nbxmpp.util import parse_websocket_uri

log = logging.getLogger("nbxmpp.addresses")


class ServerAddresses(Observable):
    """
    Signals:

        resolved

    """

    def __init__(self, domain: str | None) -> None:
        Observable.__init__(self, log)

        self._domain = domain
        self._host_meta_data: bytes | None = None
        self._custom_host: CustomHostT | None = None
        self._proxy: ProxyData | None = None
        self._is_resolved = False

        self._addresses: list[ServerAddress] = [
            ServerAddress(
                domain=self._domain,
                service="xmpps-client",
                host=None,
                uri=None,
                protocol=ConnectionProtocol.TCP,
                type=ConnectionType.DIRECT_TLS,
                proxy=None,
            ),
            ServerAddress(
                domain=self._domain,
                service="xmpp-client",
                host=None,
                uri=None,
                protocol=ConnectionProtocol.TCP,
                type=ConnectionType.START_TLS,
                proxy=None,
            ),
            ServerAddress(
                domain=self._domain,
                service="xmpp-client",
                host=None,
                uri=None,
                protocol=ConnectionProtocol.TCP,
                type=ConnectionType.PLAIN,
                proxy=None,
            ),
        ]

        self._fallback_addresses: list[ServerAddress] = [
            ServerAddress(
                domain=self._domain,
                service=None,
                host="%s:%s" % (self._domain, 5222),
                uri=None,
                protocol=ConnectionProtocol.TCP,
                type=ConnectionType.START_TLS,
                proxy=None,
            ),
            ServerAddress(
                domain=self._domain,
                service=None,
                host="%s:%s" % (self._domain, 5222),
                uri=None,
                protocol=ConnectionProtocol.TCP,
                type=ConnectionType.PLAIN,
                proxy=None,
            ),
        ]

    @property
    def domain(self) -> str | None:
        return self._domain

    @property
    def is_resolved(self) -> bool:
        return self._is_resolved

    def set_host_meta_data(self, data: bytes | None) -> None:
        self._host_meta_data = data

    def resolve(self) -> None:
        if self._is_resolved:
            self._on_request_resolved()
            return

        if self._proxy is not None:
            # Let the proxy resolve the domain
            self._on_request_resolved()
            return

        if self._custom_host is not None:
            self._on_request_resolved()
            return

        self._resolve_alternatives()

    def _resolve_alternatives(self) -> None:
        if not self._host_meta_data:
            log.info("No host meta data to process")
            self._on_request_resolved()
            return

        try:
            uri = parse_websocket_uri(self._host_meta_data.decode())
        except Exception as error:
            log.info("Error parsing websocket uri: %s", error)
            self._on_request_resolved()
            return

        if uri.startswith("wss"):
            type_ = ConnectionType.DIRECT_TLS
        elif uri.startswith("ws"):
            type_ = ConnectionType.PLAIN
        else:
            log.warning("Invalid websocket uri: %s", uri)
            self._on_request_resolved()
            return

        addr = ServerAddress(
            domain=self._domain,
            service=None,
            host=None,
            uri=uri,
            protocol=ConnectionProtocol.WEBSOCKET,
            type=type_,
            proxy=None,
        )
        self._addresses.append(addr)

        self._on_request_resolved()

    def cancel_resolve(self) -> None:
        self.remove_subscriptions()

    def set_custom_host(self, address: CustomHostT | None) -> None:
        # Set a custom host, overwrites all other addresses
        self._custom_host = address
        if address is None:
            return

        host_or_uri, protocol, type_ = address
        if protocol == ConnectionProtocol.WEBSOCKET:
            host, uri = None, host_or_uri
        else:
            host, uri = host_or_uri, None

        self._fallback_addresses = []
        self._addresses = [
            ServerAddress(
                domain=self._domain,
                service=None,
                host=host,
                uri=uri,
                protocol=protocol,
                type=type_,
                proxy=None,
            )
        ]

    def set_proxy(self, proxy: ProxyData | None) -> None:
        self._proxy = proxy

    def _on_request_resolved(self) -> None:
        self._is_resolved = True
        self.notify("resolved")
        self.remove_subscriptions()

    def get_next_address(
        self,
        allowed_types: list[ConnectionType],
        allowed_protocols: list[ConnectionProtocol],
    ) -> Iterator[ServerAddress]:
        """
        Selects next address
        """

        for addr in self._filter_allowed(
            self._addresses, allowed_types, allowed_protocols
        ):
            yield self._assure_proxy(addr)

        for addr in self._filter_allowed(
            self._fallback_addresses, allowed_types, allowed_protocols
        ):
            yield self._assure_proxy(addr)

        raise NoMoreAddresses

    def _assure_proxy(self, addr: ServerAddress) -> ServerAddress:
        if self._proxy is None:
            return addr

        if addr.protocol == ConnectionProtocol.TCP:
            return addr._replace(proxy=self._proxy)

        return addr

    def _filter_allowed(
        self,
        addresses: list[ServerAddress],
        allowed_types: list[ConnectionType],
        allowed_protocols: list[ConnectionProtocol],
    ) -> Iterator[ServerAddress]:
        if self._proxy is not None and self._proxy.type != "direct":
            addresses = filter(lambda addr: addr.host is not None, addresses)

        addresses = filter(lambda addr: addr.type in allowed_types, addresses)
        addresses = filter(lambda addr: addr.protocol in allowed_protocols, addresses)
        return addresses

    def __str__(self) -> str:
        addresses = self._addresses + self._fallback_addresses
        return "\n".join([str(addr) for addr in addresses])


class NoMoreAddresses(Exception):
    pass
